/** 
 * W-DATA format
 * @author Gabriel Wlazlowski, Warsaw University of Technology, 2020
 * 
 * Tool for creating metadafile for dpca files
 * */

// wdata lib
#include "wdata.h"

// other libs
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h> 
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#define PCA_PRECISION 8

/**
 * return
 * 0 - OK, else - problem
 * */
int read_measurement_file_header(const char * file_name, 
                                        int *nx, int *ny, int *nz, 
                                        double *dx, double *dy, double *dz,
                                        double *eF, double *t0, double *dt,
                                        int *number_of_entries
                                       )
{
    
    int fd ; /* file descriptor for handling the file or device */
    mode_t fd_mode = S_IRUSR | S_IRGRP | S_IROTH; /* S_IRWXU ; S_IRGRP, S_IRWXG ; S_IROTH , S_IRWXO; etc. */
    if ( ( fd = open( file_name , O_RDONLY  , fd_mode ) ) == -1 ) return -1;
    
    int werr=0;
    long int bytes_read ;
    int prec;
    
#define mio_hentry_read(var,type,werrval) \
    if (werr==0 && ( bytes_read = read( fd , ( void * ) var , sizeof( type ) ) ) != sizeof( type ) ) werr=werrval;

    mio_hentry_read(&prec,int,-2);
    if(werr==0 && prec!=PCA_PRECISION) werr=-3;
    mio_hentry_read(nx,int,-4);
    mio_hentry_read(ny,int,-5);
    mio_hentry_read(nz,int,-6);
    mio_hentry_read(dx,double,-7);
    mio_hentry_read(dy,double,-8);
    mio_hentry_read(dz,double,-9);
    mio_hentry_read(eF,double,-10);
    mio_hentry_read(t0,double,-11);
    mio_hentry_read(dt,double,-12);
    mio_hentry_read(number_of_entries,int,-13);
    
    // Close file
    close(fd);
    
    return werr;    
}

int main( int argc , char ** argv ) 
{
    printf("DPCA -> WDATA CONVERTER\n");
    
    if(argc!=2)
    {
        printf("Usage: %s prefix\n", argv[0]);
        printf("\tprefix - prefix of dpca files\n");
        return 0;
    }
    
    char ctmp1[512];
    strcpy(ctmp1, argv[1]);
    char *indir = dirname(ctmp1);
    printf("WORKING DIR: `%s`\n", indir);
    
    char file_name[1024];
    char file_wtxt[1024];
    sprintf(file_name, "%s_density_a.dpca", argv[1]);
    sprintf(file_wtxt, "%s.wtxt", argv[1]);
    
    printf("READING HEADER OF `%s`\n", file_name);
    
    // read dpca header needed for creation of metadata
    wdata_metadata md;
    
    double eF;
    read_measurement_file_header(file_name, &md.NX, &md.NY, &md.NZ, &md.DX, &md.DY, &md.DZ, &eF, &md.t0, &md.dt, &md.cycles);

    // set other parameters
    strcpy(md.prefix,basename(argv[1]));
    md.datadim=3;
    
    // add variables
    wdata_variable vdensity_a = {"density_a", "real", "none", "dpca"};
    wdata_add_variable(&md, &vdensity_a);
    
    wdata_variable vdensity_b = {"density_b", "real", "none", "dpca"};
    wdata_add_variable(&md, &vdensity_b);
    
    wdata_variable vdelta = {"delta", "complex", "none", "dpca"};
    wdata_add_variable(&md, &vdelta);
    
    wdata_variable vcurrent_a = {"current_a", "vector", "none", "dpca"};
    wdata_add_variable(&md, &vcurrent_a);
    
    wdata_variable vcurrent_b = {"current_b", "vector", "none", "dpca"};
    wdata_add_variable(&md, &vcurrent_b);
    
    wdata_setconst(&md, "eF", eF);
    
    printf("--------------------------------\n");
    wdata_print_metadata(&md, stdout);
    printf("--------------------------------\n");
    
   
    FILE * wtxt = fopen(file_wtxt, "w");
    fprintf(wtxt, "# Generated by: %s\n", argv[0]);
    wdata_print_metadata(&md, wtxt);
    fclose(wtxt);
    printf("GENERATED `%s`\n", file_wtxt);
    
    return 0;
}
